@@ -414,23 +414,23 @@ class Highlight(RestCall):
     """ **************************************************************************************************************
                                             Cloud Ready Data 
     ************************************************************************************************************** """
-    def get_cloud_detail(self,app_name:str)->DataFrame:
-        """Highlight cloud ready data
+    # def get_cloud_detail(self,app_name:str)->DataFrame:
+    #     """Highlight cloud ready data
 
-        Args:
-            app_name (str): name of the application
+    #     Args:
+    #         app_name (str): name of the application
 
-        Returns:
-            DataFrame: flattened version of the Highlight cloud ready data
-        """
-        try:
-            return json_normalize(self._get_metrics(app_name)['cloudReadyDetail'],['cloudReadyDetails'],meta=['technology','cloudReadyScan'])
-        except KeyError as ke:
-            self.warning(f'{app_name} has no Cloud Ready Data')
-            return None
+    #     Returns:
+    #         DataFrame: flattened version of the Highlight cloud ready data
+    #     """
+    #     try:
+    #         return json_normalize(self._get_metrics(app_name)['cloudReadyDetail'],['cloudReadyDetails'],meta=['technology','cloudReadyScan'])
+    #     except KeyError as ke:
+    #         self.warning(f'{app_name} has no Cloud Ready Data')
+    #         return None
 
     """ **************************************************************************************************************
-                                            Cloud Ready Data 
+                                            Green Detail Data 
     ************************************************************************************************************** """
     def get_green_detail(self,app_name:str)->DataFrame:
         """Highlight green impact data
@@ -447,6 +447,27 @@ class Highlight(RestCall):
             self.warning(f'{app_name} has no Green Impact Data')
             return None
 
+    _cloud={}
+    def get_cloud_detail(self,app_name:str)->DataFrame:
+        if app_name in Highlight._cloud:
+            return Highlight._cloud[app_name]
+        
+
+        columns = ['technology','cloudRequirement.display','cloudRequirement.ruleType','roadblocks','cloudEffort','cloudRequirement.criticality','files']
+
+        cloud_data = json_normalize(self._get_metrics(app_name)['cloudReadyDetail'])
+        rslt = DataFrame()
+        for index,row in cloud_data.iterrows():
+            tech = row['technology']
+            db = json_normalize(row['cloudReadyDetails'])
+            db['technology']=tech
+            db = db[columns]
+            rslt = concat([rslt,db],ignore_index=True)
+            pass
+
+        Highlight._cloud[app_name]=rslt
+        return rslt
+
     """ **************************************************************************************************************
                                             General Metrics Data 
     ************************************************************************************************************** """
